# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to cycler run data structures"""

import os
import unittest

from beep.structure.arbin import ArbinDatapath
from beep.structure.cli import auto_load, auto_load_processed
from beep.tests.constants import TEST_FILE_DIR


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.arbin_file = os.path.join(
            TEST_FILE_DIR, "2017-12-04_4_65C-69per_6C_CH29.csv"
        )

        self.processed_file = os.path.join(
            TEST_FILE_DIR, "2017-06-30_2C-10per_6C_CH10_structure.json"
        )

        self.processed_maccor_file = os.path.join(
            TEST_FILE_DIR, "PredictionDiagnostics_000132_00004C_structure.json"
        )

    # todo: could be more comprehensive
    # based on PCRT.test_auto_load
    def test_auto_load(self):
        dp = auto_load(self.arbin_file)
        self.assertIsInstance(dp, ArbinDatapath)

    def test_auto_load_processed(self):
        dp = auto_load_processed(self.processed_file)
        self.assertIsInstance(dp, ArbinDatapath)
        self.assertIsNotNone(dp.structured_summary)

    def test_auto_load_maccor(self):
        dp = auto_load_processed(self.processed_maccor_file)
        self.assertEqual(dp.paths["structured"], self.processed_maccor_file)



if __name__ == "__main__":
    unittest.main()



import os

import pandas as pd

from beep import PROTOCOL_PARAMETERS_DIR
from beep.features import featurizer_helpers
from beep.features.base import BEEPFeaturizer
from beep.features.intracell_analysis_v2 import IntracellAnalysisV2


DEFAULT_CELL_INFO_DIR = os.path.join(PROTOCOL_PARAMETERS_DIR, "intracell_info")


class IntracellCyclesV2(BEEPFeaturizer):
    """
    Object corresponding to the fitted material parameters of the cell. Material parameters
    are determined by using high resolution half cell data to fit full cell dQdV curves. Rows
    of the output dataframe correspond to each of the diagnostics throughout the life of the
    cell.
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    DEFAULT_HYPERPARAMETERS = {
        "diagnostic_cycle_type": 'rpt_0.2C',
        "step_type": 0,
        # Paths for anode files should be absolute
        # Defaults are for the specified names in the current dir
        "anode_file": os.path.join(
            DEFAULT_CELL_INFO_DIR,
            'anode_test.csv'
        ),
        "cathode_file": os.path.join(
            DEFAULT_CELL_INFO_DIR,
            'cathode_test.csv'
        ),
    }

    def validate(self):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations

        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        val, msg = featurizer_helpers.check_diagnostic_validation(self.datapath)
        if val:
            conditions = []

            # Ensure overlap of cycle indices above threshold and matching cycle type
            eol_cycle_index_list = self.datapath.diagnostic_summary[
                (self.datapath.diagnostic_summary.cycle_type == self.hyperparameters["diagnostic_cycle_type"]) &
                (self.datapath.diagnostic_summary.discharge_capacity > IntracellAnalysisV2.THRESHOLD)
                ].cycle_index.to_list()
            if not eol_cycle_index_list:
                return False, "Overlap of cycle indices not above threshold for matching cycle type"

            conditions.append(
                any(
                    [
                        "rpt" in x
                        for x in self.datapath.diagnostic_summary.cycle_type.unique()
                    ]
                )
            )

            if all(conditions):
                return True, None
            else:
                return False, "Insufficient RPT cycles in diagnostic"
        else:
            return val, msg

    def create_features(self):
        """
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            self.hyperparameters (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.
            cell_info_path (str): Root directory for cell half cell data

        Returns:
             (pd.DataFrame) containing the cell material parameters as a function of cycle index
        """
        ia = IntracellAnalysisV2(
            self.hyperparameters["cathode_file"],
            self.hyperparameters["anode_file"],
            cycle_type=self.hyperparameters["diagnostic_cycle_type"],
            step_type=self.hyperparameters["step_type"]
        )

        # (cell_init_aligned, cell_init_profile, PE_matched, NE_matched) = ia.intracell_wrapper_init(
        #     self.datapath,
        # )

        eol_cycle_index_list = self.datapath.diagnostic_summary[
            (self.datapath.diagnostic_summary.cycle_type == ia.cycle_type) &
            (self.datapath.diagnostic_summary.discharge_capacity > ia.THRESHOLD)
            ].cycle_index.to_list()

        # initialize dicts before for loop
        dataset_dict_of_cell_degradation_path = dict()
        real_cell_dict_of_profiles = dict()
        for i, cycle_index in enumerate(eol_cycle_index_list):
            loss_dict, profiles_dict = ia.intracell_values_wrapper_ah(cycle_index,
                                                                      self.datapath
                                                                      )
            dataset_dict_of_cell_degradation_path.update(loss_dict)
            real_cell_dict_of_profiles.update(profiles_dict)

        degradation_df = pd.DataFrame(dataset_dict_of_cell_degradation_path,
                                      index=['rmse_error', 'LLI_opt', 'Q_pe_opt', 'Q_ne_opt', 'x_NE_2',
                                             'pe_voltage_FC4p2V', 'pe_voltage_FC4p1V', 'pe_voltage_FC4p0V',
                                             'pe_voltage_FC3p9V', 'pe_voltage_FC3p8V', 'pe_voltage_FC3p7V',
                                             'pe_voltage_FC3p6V', 'pe_voltage_FC3p5V', 'pe_voltage_FC3p4V',
                                             'pe_voltage_FC3p3V', 'pe_voltage_FC3p2V', 'pe_voltage_FC3p1V',
                                             'pe_voltage_FC3p0V', 'pe_voltage_FC2p9V', 'pe_voltage_FC2p8V',
                                             'pe_voltage_FC2p7V',
                                             'pe_soc_FC4p2V', 'pe_soc_FC4p1V', 'pe_soc_FC4p0V', 'pe_soc_FC3p9V',
                                             'pe_soc_FC3p8V', 'pe_soc_FC3p7V',
                                             'pe_soc_FC3p6V', 'pe_soc_FC3p5V', 'pe_soc_FC3p4V', 'pe_soc_FC3p3V',
                                             'pe_soc_FC3p2V', 'pe_soc_FC3p1V',
                                             'pe_soc_FC3p0V', 'pe_soc_FC2p9V', 'pe_soc_FC2p8V', 'pe_soc_FC2p7V',
                                             'ne_voltage_FC4p2V', 'ne_voltage_FC4p1V', 'ne_voltage_FC4p0V',
                                             'ne_voltage_FC3p9V', 'ne_voltage_FC3p8V', 'ne_voltage_FC3p7V',
                                             'ne_voltage_FC3p6V', 'ne_voltage_FC3p5V', 'ne_voltage_FC3p4V',
                                             'ne_voltage_FC3p3V', 'ne_voltage_FC3p2V', 'ne_voltage_FC3p1V',
                                             'ne_voltage_FC3p0V', 'ne_voltage_FC2p9V', 'ne_voltage_FC2p8V',
                                             'ne_voltage_FC2p7V',
                                             'ne_soc_FC4p2V', 'ne_soc_FC4p1V', 'ne_soc_FC4p0V', 'ne_soc_FC3p9V',
                                             'ne_soc_FC3p8V', 'ne_soc_FC3p7V',
                                             'ne_soc_FC3p6V', 'ne_soc_FC3p5V', 'ne_soc_FC3p4V', 'ne_soc_FC3p3V',
                                             'ne_soc_FC3p2V', 'ne_soc_FC3p1V',
                                             'ne_soc_FC3p0V', 'ne_soc_FC2p9V', 'ne_soc_FC2p8V', 'ne_soc_FC2p7V',
                                             'Q_fc', 'Q_pe', 'Q_ne', 'Q_li']).T
        self.features = degradation_df


class IntracellFeaturesV2(IntracellCyclesV2):
    """
    Object corresponding to the fitted material parameters of the cell. Material parameters
    are determined by using high resolution half cell data to fit full cell dQdV curves. The
    parameters from the first and second diagnostics are used as the feature values.
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    def create_features(self):
        """
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.
            cell_info_path (str): Root directory for cell half cell data


        Returns:
             (pd.DataFrame) containing the cell material parameters for the first and second diagnotics
                as a single row dataframe
        """

        ia = IntracellAnalysisV2(
            self.hyperparameters["cathode_file"],
            self.hyperparameters["anode_file"],
            cycle_type=self.hyperparameters["diagnostic_cycle_type"],
            step_type=self.hyperparameters["step_type"]
        )

        # (cell_init_aligned, cell_init_profile, PE_matched, NE_matched) = ia.intracell_wrapper_init(
        #     self.datapath,
        # )

        eol_cycle_index_list = self.datapath.diagnostic_summary[
            (self.datapath.diagnostic_summary.cycle_type == ia.cycle_type) &
            (self.datapath.diagnostic_summary.discharge_capacity > ia.THRESHOLD)
            ].cycle_index.to_list()

        # # initializations before for loop
        dataset_dict_of_cell_degradation_path = dict()
        real_cell_dict_of_profiles = dict()
        for i, cycle_index in enumerate(eol_cycle_index_list[0:2]):
            loss_dict, profiles_dict = ia.intracell_values_wrapper_ah(cycle_index,
                                                                      self.datapath
                                                                      )
            dataset_dict_of_cell_degradation_path.update(loss_dict)
            real_cell_dict_of_profiles.update(profiles_dict)

        degradation_df = pd.DataFrame(dataset_dict_of_cell_degradation_path,
                                      index=['rmse_error', 'LLI_opt', 'Q_pe_opt', 'Q_ne_opt', 'x_NE_2',
                                             'pe_voltage_FC4p2V', 'pe_voltage_FC4p1V', 'pe_voltage_FC4p0V',
                                             'pe_voltage_FC3p9V', 'pe_voltage_FC3p8V', 'pe_voltage_FC3p7V',
                                             'pe_voltage_FC3p6V', 'pe_voltage_FC3p5V', 'pe_voltage_FC3p4V',
                                             'pe_voltage_FC3p3V', 'pe_voltage_FC3p2V', 'pe_voltage_FC3p1V',
                                             'pe_voltage_FC3p0V', 'pe_voltage_FC2p9V', 'pe_voltage_FC2p8V',
                                             'pe_voltage_FC2p7V',
                                             'pe_soc_FC4p2V', 'pe_soc_FC4p1V', 'pe_soc_FC4p0V', 'pe_soc_FC3p9V',
                                             'pe_soc_FC3p8V', 'pe_soc_FC3p7V',
                                             'pe_soc_FC3p6V', 'pe_soc_FC3p5V', 'pe_soc_FC3p4V', 'pe_soc_FC3p3V',
                                             'pe_soc_FC3p2V', 'pe_soc_FC3p1V',
                                             'pe_soc_FC3p0V', 'pe_soc_FC2p9V', 'pe_soc_FC2p8V', 'pe_soc_FC2p7V',
                                             'ne_voltage_FC4p2V', 'ne_voltage_FC4p1V', 'ne_voltage_FC4p0V',
                                             'ne_voltage_FC3p9V', 'ne_voltage_FC3p8V', 'ne_voltage_FC3p7V',
                                             'ne_voltage_FC3p6V', 'ne_voltage_FC3p5V', 'ne_voltage_FC3p4V',
                                             'ne_voltage_FC3p3V', 'ne_voltage_FC3p2V', 'ne_voltage_FC3p1V',
                                             'ne_voltage_FC3p0V', 'ne_voltage_FC2p9V', 'ne_voltage_FC2p8V',
                                             'ne_voltage_FC2p7V',
                                             'ne_soc_FC4p2V', 'ne_soc_FC4p1V', 'ne_soc_FC4p0V', 'ne_soc_FC3p9V',
                                             'ne_soc_FC3p8V', 'ne_soc_FC3p7V',
                                             'ne_soc_FC3p6V', 'ne_soc_FC3p5V', 'ne_soc_FC3p4V', 'ne_soc_FC3p3V',
                                             'ne_soc_FC3p2V', 'ne_soc_FC3p1V',
                                             'ne_soc_FC3p0V', 'ne_soc_FC2p9V', 'ne_soc_FC2p8V', 'ne_soc_FC2p7V',
                                             'Q_fc', 'Q_pe', 'Q_ne', 'Q_li']).T

        diag_0_names = ["diag_0_" + name for name in degradation_df.columns]
        diag_1_names = ["diag_1_" + name for name in degradation_df.columns]
        values = {0: degradation_df.iloc[0].tolist() + degradation_df.iloc[1].tolist()}
        features_df = pd.DataFrame(values, index=diag_0_names+diag_1_names).T
        self.features = features_df



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to the fitting functions that extract cell parameters from dis profiles"""

import os
import unittest
import pandas as pd
from beep.structure.cli import auto_load_processed
from beep.features.intracell_analysis_v2 import IntracellAnalysisV2
from beep.features.intracell_losses_v2 import IntracellCyclesV2, IntracellFeaturesV2

from beep.tests.constants import TEST_FILE_DIR


class IntracellAnalysisV2Test(unittest.TestCase):
    def setUp(self):
        run = os.path.join(TEST_FILE_DIR,
                           'PreDiag_000220_00005E_structure_omit.json')
        self.cell_struct = auto_load_processed(run)

    def test_intracell_wrappers(self):
        ia = IntracellAnalysisV2(
            os.path.join(TEST_FILE_DIR, 'cathode_clean_cc_charge_exptl_aligned.csv'),
            os.path.join(TEST_FILE_DIR, 'anode_secondMeasure_clean_cc_charge_exptl_aligned.csv'),
            cycle_type='rpt_0.2C',
            error_type='dVdQ'
        )

        cycle_index = 3

        loss_dict, profiles_dict = ia.intracell_values_wrapper_ah(cycle_index, self.cell_struct)

        loss_df = pd.DataFrame(loss_dict, index=['rmse_error', 'LLI_opt', 'Q_pe_opt', 'Q_ne_opt', 'x_NE_2',
                                                 'pe_voltage_FC4p2V', 'pe_voltage_FC4p1V', 'pe_voltage_FC4p0V',
                                                 'pe_voltage_FC3p9V', 'pe_voltage_FC3p8V', 'pe_voltage_FC3p7V',
                                                 'pe_voltage_FC3p6V', 'pe_voltage_FC3p5V', 'pe_voltage_FC3p4V',
                                                 'pe_voltage_FC3p3V', 'pe_voltage_FC3p2V', 'pe_voltage_FC3p1V',
                                                 'pe_voltage_FC3p0V', 'pe_voltage_FC2p9V', 'pe_voltage_FC2p8V',
                                                 'pe_voltage_FC2p7V',
                                                 'pe_soc_FC4p2V', 'pe_soc_FC4p1V', 'pe_soc_FC4p0V', 'pe_soc_FC3p9V',
                                                 'pe_soc_FC3p8V', 'pe_soc_FC3p7V',
                                                 'pe_soc_FC3p6V', 'pe_soc_FC3p5V', 'pe_soc_FC3p4V', 'pe_soc_FC3p3V',
                                                 'pe_soc_FC3p2V', 'pe_soc_FC3p1V',
                                                 'pe_soc_FC3p0V', 'pe_soc_FC2p9V', 'pe_soc_FC2p8V', 'pe_soc_FC2p7V',
                                                 'ne_voltage_FC4p2V', 'ne_voltage_FC4p1V', 'ne_voltage_FC4p0V',
                                                 'ne_voltage_FC3p9V', 'ne_voltage_FC3p8V', 'ne_voltage_FC3p7V',
                                                 'ne_voltage_FC3p6V', 'ne_voltage_FC3p5V', 'ne_voltage_FC3p4V',
                                                 'ne_voltage_FC3p3V', 'ne_voltage_FC3p2V', 'ne_voltage_FC3p1V',
                                                 'ne_voltage_FC3p0V', 'ne_voltage_FC2p9V', 'ne_voltage_FC2p8V',
                                                 'ne_voltage_FC2p7V',
                                                 'ne_soc_FC4p2V', 'ne_soc_FC4p1V', 'ne_soc_FC4p0V', 'ne_soc_FC3p9V',
                                                 'ne_soc_FC3p8V', 'ne_soc_FC3p7V',
                                                 'ne_soc_FC3p6V', 'ne_soc_FC3p5V', 'ne_soc_FC3p4V', 'ne_soc_FC3p3V',
                                                 'ne_soc_FC3p2V', 'ne_soc_FC3p1V',
                                                 'ne_soc_FC3p0V', 'ne_soc_FC2p9V', 'ne_soc_FC2p8V', 'ne_soc_FC2p7V',
                                                 'Q_fc', 'Q_pe', 'Q_ne', 'Q_li']).T

        self.assertAlmostEqual(loss_df['Q_pe'].iloc[0], 4.971001250967573, 3)
        self.assertAlmostEqual(loss_df['Q_ne'].iloc[0], 5.077560763314145, 3)
        self.assertAlmostEqual(loss_df['Q_li'].iloc[0], 4.72058639695313, 3)
        self.assertAlmostEqual(loss_df['pe_voltage_FC4p2V'].iloc[0], 4.2569909388206595, 3)
        self.assertAlmostEqual(loss_df['pe_voltage_FC2p7V'].iloc[0], 3.627263943506632, 3)
        self.assertAlmostEqual(loss_df['ne_voltage_FC4p2V'].iloc[0], 0.05688501257876296, 5)
        self.assertAlmostEqual(loss_df['ne_voltage_FC2p7V'].iloc[0], 0.7952915950646048, 5)
        self.assertAlmostEqual(loss_df['pe_soc_FC4p2V'].iloc[0], 0.9646302250803859, 5)
        self.assertAlmostEqual(loss_df['pe_soc_FC2p7V'].iloc[0], 0.05144694533762058, 5)
        self.assertAlmostEqual(loss_df['ne_soc_FC4p2V'].iloc[0], 0.895068205666317, 5)
        self.assertAlmostEqual(loss_df['ne_soc_FC2p7V'].iloc[0], 0.0010493179433368257, 5)


class IntracellFeaturesTestV2(unittest.TestCase):
    def setUp(self):
        run_path = os.path.join(TEST_FILE_DIR, 'PreDiag_000220_00005E_structure_omit.json')
        self.datapath = auto_load_processed(run_path)
        cathode_file = os.path.join(TEST_FILE_DIR, 'cathode_clean_cc_charge_exptl_aligned.csv')
        anode_file = os.path.join(TEST_FILE_DIR, 'anode_secondMeasure_clean_cc_charge_exptl_aligned.csv')

        self.params = {
            'diagnostic_cycle_type': 'rpt_0.2C',
            'step_type': 0,
            "anode_file": anode_file,
            "cathode_file": cathode_file
        }

    def test_IntracellCycles(self):
        featurizer = IntracellCyclesV2(self.datapath, self.params)
        featurizer.create_features()
        X = featurizer.features
        self.assertEqual(X.shape, (2, 73))
        self.assertAlmostEqual(X["Q_li"].iloc[0], 4.743450821877655, 5)
        self.assertAlmostEqual(X["Q_ne"].iloc[1], 5.101834164537508, 3)

    def test_IntracellFeatures(self):
        featurizer = IntracellFeaturesV2(self.datapath, self.params)
        featurizer.create_features()
        X = featurizer.features
        self.assertEqual(X.shape, (1, 146))
        self.assertAlmostEqual(X["diag_0_LLI_opt"].iloc[0], 0.1929531780384086, 5)
        self.assertAlmostEqual(X["diag_1_LLI_opt"].iloc[0], 0.2080220199958258, 3)

    def test_validation(self):
        # Modify datapath_run to be invalid
        mask = self.datapath.diagnostic_summary.cycle_type == "rpt_0.2C"
        self.datapath.diagnostic_summary.loc[mask, "discharge_capacity"] = 0

        featurizer = IntracellFeaturesV2(self.datapath, self.params)
        val, msg = featurizer.validate()
        self.assertFalse(val)



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Classes for advanced or specialized validation procedures on cycler data.

Validation classes are designed to work with datapaths, meaning
column names of input dataframes are standardized according to the
validation schema.

"""
import os

import numpy as np
from monty.serialization import loadfn

from beep import VALIDATION_SCHEMA_DIR

DEFAULT_ARBIN_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-arbin-lfp.yaml")
DEFAULT_EIS_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-maccor-eis.yaml")
PROJECT_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-projects.yaml")


class SimpleValidator:
    """
    Lightweight class that does Dataframe-based, as
    opposed to dictionary based validation

    Note that schemas here are made to be identical to cerberus
    schemas and should support similar syntax, e. g.

    {COLUMN_NAME:
        {schema:
            type: TYPE_IN_COLUMN, [int, float, str, object]
            max: MAX_VALUE_IN_COLUMN,
            min: MIN_VALUE_IN_COLUMN
         type: list
        }
    }

    Note that the COLUMN_NAME.type key above is ignored, but
    COLUMN_NAME.schema.type is used.

    The only schema keys that are supported at this time are
    max, min, and type.

    Typing is compared using the key-mapping by rule defined
    by the ALLOWED_TYPES_BY_RULE attribute defined below.
    Supported type rules include "integer", "float", "numeric",
    and "string".  Note that type-checking for this class is
    not equivalent to checking types, and may involve custom
    logic which is defined in the check_type method below.
    """

    def __init__(self, schema_filename=DEFAULT_ARBIN_SCHEMA):
        """
        Args:
            schema (dict): Schema to validate against.
        """
        self.schema = loadfn(schema_filename)
        self.validation_records = None

    @staticmethod
    def check_type(df, type_rule):
        """
        Method to check type of input dataframe.

        Args:
            df (pandas.Dataframe): DataFrame.
            type_rule (str): string corresponding to type_rule
                to check, supported type rules are:
                integer: checks for numeric values which are
                    equal to their rounded values

        Returns:
            bool: valid
            str: verbose description of reason
        """
        if type_rule not in ["integer", "float", "numeric", "string"]:
            raise ValueError(
                "type_rule {} not supported, please choose one "
                "of integer, float, numeric, or string"
            )
        # Integer: Check residual from rounding
        if type_rule == "integer":
            nonint_indices = np.arange(len(df))[(df != np.round(df))]
            if nonint_indices.size > 0:
                value = df.iloc[nonint_indices[0]]
                return (
                    False,
                    "integer type check failed at index {} with value {}".format(
                        nonint_indices[0], value
                    ),
                )
        # Float: just check numpy dtyping
        elif type_rule == "float":
            if not np.issubdtype(df.dtype, np.floating):
                return False, "float type check failed, type is {}".format(df.dtype)

        # Numeric: check numpy number dtyping
        elif type_rule == "numeric":
            if not np.issubdtype(df.dtype, np.number):
                return False, "number type check failed, type is {}".format(df.dtype)

        # String: check string/unicode subdtype
        elif type_rule == "string":
            if not (
                np.issubdtype(df.dtype, np.object_)
                or np.issubdtype(df.dtype, np.unicode_)
            ):
                return False, "string type check failed, type is {}".format(df.dtype)
        return True, ""

    def validate(self, dataframe):
        """
        Method to run the validation on everything, and report
        the results, i. e. which columns are inconsistent with
        the schema.

        Args:
            dataframe (pandas.DataFrame): dataframe to be validated.

        Returns:
            dict: report corresponding to each validation
            str: reason for report validation failure, empty string on report
                validation success
        """
        dataframe = dataframe.rename(str.lower, axis="columns")
        for column_name, value in self.schema.items():
            column_schema = value["schema"]
            max_at_least_rule = column_schema.get("max_at_least")
            min_is_below_rule = column_schema.get("min_is_below")
            max_rule = column_schema.get("max")
            min_rule = column_schema.get("min")
            type_rule = column_schema.get("type")
            monotonic_rule = column_schema.get("monotonic")

            # Check type
            if type_rule is not None:
                validity, reason = self.check_type(
                    dataframe[column_name], type_rule=type_rule
                )
                if not validity:
                    reason = "Column {}: {}".format(column_name, reason)
                    return validity, reason

            # Check max
            if max_rule is not None:
                comp = np.where(dataframe[column_name] > max_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} is higher than allowed max {} at index {}: "
                        "value={}".format(column_name, max_rule, index, value)
                    )
                    return False, reason

            # Check min
            if min_rule is not None:
                comp = np.where(dataframe[column_name] < min_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} is lower than allowed min {} at index {}:"
                        "value={}".format(column_name, min_rule, index, value)
                    )
                    return False, reason

            # Check a maximum value is at least above a threshold
            if max_at_least_rule is not None:
                comp = np.where(dataframe[column_name].max() < max_at_least_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} needs to reach at least {} for processing, instead found:"
                        "value={}".format(column_name, max_at_least_rule, value)
                    )
                    return False, reason

            # Check a minimum value is below above a threshold
            if min_is_below_rule is not None:
                comp = np.where(dataframe[column_name].min() > min_is_below_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} needs to reach under {} for processing, instead found:"
                        "value={}".format(column_name, max_at_least_rule, value)
                    )
                    return False, reason

            if monotonic_rule == 'increasing':
                diff_series = dataframe[column_name].diff().dropna()
                if len(diff_series[diff_series < 0]) > 0:
                    reason = (
                        "{} needs to be monotonically increasing for processing".format(column_name)
                    )
                    return False, reason

        return True, ""


class BEEPValidationError(Exception):
    """Custom error to raise when validation fails"""



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from scipy.spatial import distance
from scipy.optimize import differential_evolution


class IntracellAnalysisV2:
    # IA constants
    FC_UPPER_VOLTAGE = 4.20
    FC_LOWER_VOLTAGE = 2.70
    NE_UPPER_VOLTAGE = 0.01
    NE_LOWER_VOLTAGE = 1.50
    PE_UPPER_VOLTAGE = 4.30
    PE_LOWER_VOLTAGE = 2.86
    THRESHOLD = 4.84 * 0.0

    def __init__(self,
                 pe_pristine_file,
                 ne_pristine_file,
                 cycle_type='rpt_0.2C',
                 step_type=0,
                 error_type='V-Q',
                 ne_2pos_file=None,
                 ne_2neg_file=None
                 ):
        """
        Invokes the cell electrode analysis class. This is a class designed to fit the cell and electrode
        parameters in order to determine changes of electrodes within the full cell from only full cell cycling data.
        Args:
            pe_pristine_file (str): file name for the half cell data of the pristine (uncycled) positive
                electrode
            ne_pristine_file (str): file name for the half cell data of the pristine (uncycled) negative
                electrode
            cycle_type (str): type of diagnostic cycle for the fitting
            step_type (int): charge or discharge (0 for charge, 1 for discharge)
            error_type (str): defines which error metric is to be used
            ne_2neg_file (str): file name of the data for the negative component of the anode
            ne_2pos_file (str): file name of the data for the positive component of the anode
        """
        self.pe_pristine = pd.read_csv(pe_pristine_file, usecols=['SOC_aligned', 'Voltage_aligned'])
        self.ne_1_pristine = pd.read_csv(ne_pristine_file, usecols=['SOC_aligned', 'Voltage_aligned'])

        if ne_2neg_file and ne_2pos_file:
            self.ne_2_pristine_pos = pd.read_csv(ne_2pos_file)
            self.ne_2_pristine_neg = pd.read_csv(ne_2neg_file)
        else:
            self.ne_2_pristine_pos = pd.DataFrame()
            self.ne_2_pristine_neg = pd.DataFrame()

        if step_type == 0:
            self.capacity_col = 'charge_capacity'
        else:
            self.capacity_col = 'discharge_capacity'

        self.cycle_type = cycle_type
        self.step_type = step_type
        self.error_type = error_type

    def process_beep_cycle_data_for_candidate_halfcell_analysis_ah(self,
                                                                   cell_struct,
                                                                   cycle_index):
        """
        Ingests BEEP structured cycling data and cycle_index and returns
                a Dataframe of evenly spaced capacity with corresponding voltage.

        Inputs:
        cell_struct (MaccorDatapath): BEEP structured cycling data
        cycle_index (int): cycle number at which to evaluate

        Outputs:
        real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned (evenly spaced)
                and Voltage_aligned
        """

        # filter the data down to the diagnostic type of interest
        diag_type_cycles = cell_struct.diagnostic_data.loc[cell_struct.diagnostic_data['cycle_type'] == self.cycle_type]
        real_cell_candidate_charge_profile = diag_type_cycles.loc[
            (diag_type_cycles.cycle_index == cycle_index)
            & (diag_type_cycles.step_type == 0)  # step_type = 0 is charge, 1 is discharge
            & (diag_type_cycles.voltage < self.FC_UPPER_VOLTAGE)
            & (diag_type_cycles[self.capacity_col] > 0)][['voltage', 'charge_capacity']]

        # renaming capacity,voltage column
        real_cell_candidate_charge_profile['Q'] = real_cell_candidate_charge_profile['charge_capacity']

        real_cell_candidate_charge_profile['Voltage'] = real_cell_candidate_charge_profile['voltage']
        real_cell_candidate_charge_profile.drop('voltage', axis=1, inplace=True)

        # interpolate voltage along evenly spaced capacity axis
        q_vec = np.linspace(0, np.max(real_cell_candidate_charge_profile['Q']), 1001)

        real_cell_candidate_charge_profile_aligned = pd.DataFrame()
        real_cell_candidate_charge_profile_interper = interp1d(real_cell_candidate_charge_profile['Q'],
                                                               real_cell_candidate_charge_profile['Voltage'],
                                                               bounds_error=False,
                                                               fill_value=(
                                                               self.FC_LOWER_VOLTAGE, self.FC_UPPER_VOLTAGE))
        real_cell_candidate_charge_profile_aligned['Voltage_aligned'] = real_cell_candidate_charge_profile_interper(
            q_vec)

        real_cell_candidate_charge_profile_aligned['Q_aligned'] = q_vec

        return real_cell_candidate_charge_profile_aligned

    def _impose_electrode_scale(self,
                                pe_pristine=pd.DataFrame(),
                                ne_1_pristine=pd.DataFrame(),
                                ne_2_pristine_pos=pd.DataFrame(),
                                ne_2_pristine_neg=pd.DataFrame(),
                                lli=0.0, q_pe=0.0, q_ne=0.0, x_ne_2=0.0):
        """
        Scales the reference electrodes according to specified capacities and
        offsets their capacities according to lli. Blends negative electrode materials.

        Inputs:
        pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                electrode
        ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                electrode
        ne_2_pos (Dataframe): half cell data for the positive component of the anode
        ne_2_neg (Dataframe): half cell data for the negative component of the anode
        lli (float): Loss of Lithium Inventory - capacity of the misalignment between
                cathode and anode zero-capacity
        q_pe (float): capacity of the positive electrode (cathode)
        q_ne (float): capacity of the negative electrode (anode)
        x_ne_2 (float): fraction of ne_2_pristine_pos or ne_2_pristine_neg
                (positive or negative value, respectively) to ne_1_pristine

        Outputs:
        pe_degraded (Dataframe): positive electrode with imposed capacity
                scale to emulate degradation
        ne_degraded (Dataframe): negative electrode with imposed capacity
                scale and capacity offset to emulate degradation
        """
        # Blend negative electrodes
        ne_pristine = blend_electrodes(ne_1_pristine, ne_2_pristine_pos, ne_2_pristine_neg, x_ne_2)

        # rescaling pristine electrodes to q_pe and q_ne
        pe_q_scaled = pe_pristine.copy()
        pe_q_scaled['Q_aligned'] = (pe_q_scaled['SOC_aligned'] / 100) * q_pe
        ne_q_scaled = ne_pristine.copy()
        ne_q_scaled['Q_aligned'] = (ne_q_scaled['SOC_aligned'] / 100) * q_ne

        # translate pristine ne electrode with lli
        ne_q_scaled['Q_aligned'] = ne_q_scaled['Q_aligned'] + lli

        # Re-interpolate to align dataframes for differencing
        lower_q = np.min((np.min(pe_q_scaled['Q_aligned']),
                          np.min(ne_q_scaled['Q_aligned'])))
        upper_q = np.max((np.max(pe_q_scaled['Q_aligned']),
                          np.max(ne_q_scaled['Q_aligned'])))
        q_vec = np.linspace(lower_q, upper_q, 1001)

        # Actually aligning the electrode Q's
        pe_pristine_interper = interp1d(pe_q_scaled['Q_aligned'],
                                        pe_q_scaled['Voltage_aligned'], bounds_error=False)
        pe_degraded = pe_q_scaled.copy()
        pe_degraded['Q_aligned'] = q_vec
        pe_degraded['Voltage_aligned'] = pe_pristine_interper(q_vec)

        ne_pristine_interper = interp1d(ne_q_scaled['Q_aligned'],
                                        ne_q_scaled['Voltage_aligned'], bounds_error=False)
        ne_degraded = ne_q_scaled.copy()
        ne_degraded['Q_aligned'] = q_vec
        ne_degraded['Voltage_aligned'] = ne_pristine_interper(q_vec)

        # Returning pe and ne degraded on an Ah basis
        return pe_degraded, ne_degraded

    def halfcell_degradation_matching_ah(self, x, *params):
        """
        Calls underlying functions to impose degradation through electrode
        capacity scale and alignment through LLI. Modifies emulated full cell
        data to be within full cell voltage range and calibrates (zeros) capacity
        at the lowest permissible voltage. Interpolates real and emulated data onto
        a common capacity axis.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned
        Outputs:
        pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
                offset, and aligned along capacity
        ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
                offset, and aligned along capacity

        df_real_aligned (Dataframe): capacity and voltage interpolated evenly across
                capacity for the real cell data
        emulated_full_cell_aligned (Dataframe): capacity and voltage interpolated evenly
                across capacity for the emulated cell data
        """

        lli = x[0]
        q_pe = x[1]
        q_ne = x[2]
        x_ne_2 = x[3]

        (pe_pristine,
         ne_1_pristine,
         ne_2_pristine_pos,
         ne_2_pristine_neg,
         real_cell_candidate_charge_profile_aligned) = params

        # output degraded ne and pe (on a AH basis, with electrode alignment
        # (NaNs for voltage, when no capacity actually at the corresponding capacity index))
        pe_out, ne_out = self._impose_electrode_scale(pe_pristine, ne_1_pristine,
                                                      ne_2_pristine_pos, ne_2_pristine_neg,
                                                      lli, q_pe,
                                                      q_ne, x_ne_2)

        # PE - NE = full cell voltage
        emulated_full_cell_with_degradation = pd.DataFrame()
        emulated_full_cell_with_degradation['Q_aligned'] = pe_out['Q_aligned'].copy()
        emulated_full_cell_with_degradation['Voltage_aligned'] = pe_out['Voltage_aligned'] - ne_out['Voltage_aligned']

        # Replace emulated full cell values outside of voltage range with NaN
        emulated_full_cell_with_degradation['Voltage_aligned'].loc[
            emulated_full_cell_with_degradation['Voltage_aligned'] < self.FC_LOWER_VOLTAGE] = np.nan
        emulated_full_cell_with_degradation['Voltage_aligned'].loc[
            emulated_full_cell_with_degradation['Voltage_aligned'] > self.FC_UPPER_VOLTAGE] = np.nan

        # Center the emulated full cell and half cell curves onto the same Q at which the real (degraded)
        # capacity measurement started (self.FC_LOWER_VOLTAGE)
        emulated_full_cell_with_degradation_zeroed = pd.DataFrame()

        emulated_full_cell_with_degradation_zeroed['Voltage_aligned'] = emulated_full_cell_with_degradation[
            'Voltage_aligned'].copy()

        zeroing_value = emulated_full_cell_with_degradation['Q_aligned'].loc[
            np.nanargmin(emulated_full_cell_with_degradation['Voltage_aligned'])
        ]

        emulated_full_cell_with_degradation_zeroed['Q_aligned'] = \
            (emulated_full_cell_with_degradation['Q_aligned'].copy() - zeroing_value)

        pe_out_zeroed = pe_out.copy()
        pe_out_zeroed['Q_aligned'] = pe_out['Q_aligned'] - zeroing_value

        ne_out_zeroed = ne_out.copy()
        ne_out_zeroed['Q_aligned'] = ne_out['Q_aligned'] - zeroing_value

        # Interpolate full cell profiles across same Q range
        min_q = np.min(
            real_cell_candidate_charge_profile_aligned['Q_aligned'].loc[
                ~real_cell_candidate_charge_profile_aligned['Voltage_aligned'].isna()])
        max_q = np.max(
            real_cell_candidate_charge_profile_aligned['Q_aligned'].loc[
                ~real_cell_candidate_charge_profile_aligned['Voltage_aligned'].isna()])
        emulated_interper = interp1d(emulated_full_cell_with_degradation_zeroed['Q_aligned'].loc[
                                         ~emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].isna()],
                                     emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].loc[
                                         ~emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].isna()],
                                     bounds_error=False)
        real_interper = interp1d(
            real_cell_candidate_charge_profile_aligned['Q_aligned'].loc[
                ~real_cell_candidate_charge_profile_aligned['Voltage_aligned'].isna()],
            real_cell_candidate_charge_profile_aligned['Voltage_aligned'].loc[
                ~real_cell_candidate_charge_profile_aligned['Voltage_aligned'].isna()],
            bounds_error=False)

        q_vec = np.linspace(min_q, max_q, 1001)

        emulated_aligned = pd.DataFrame()
        emulated_aligned['Q_aligned'] = q_vec
        emulated_aligned['Voltage_aligned'] = emulated_interper(q_vec)

        real_aligned = pd.DataFrame()
        real_aligned['Q_aligned'] = q_vec
        real_aligned['Voltage_aligned'] = real_interper(q_vec)

        return pe_out_zeroed, ne_out_zeroed, real_aligned, emulated_aligned

    def get_dqdv_over_v_from_degradation_matching_ah(self, x, *params):
        """
        This function imposes degradation scaling ,then outputs the dqdv representation of the emulated cell data.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
        pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
                offset, and aligned along capacity
        ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
                offset, and aligned along capacity
        dq_dv_over_v_real (Dataframe): dqdv across voltage for the real cell data
        dq_dv_over_v_emulated (Dataframe): dqdv across voltage for the emulated cell data
        df_real_interped (Dataframe): capacity and voltage interpolated evenly across
                capacity for the real cell data
        emulated_full_cell_interped (Dataframe): capacity and voltage interpolated evenly
                across capacity for the emulated cell data
        """

        pe_out_zeroed, ne_out_zeroed, df_real_interped, emulated_full_cell_interped = \
            self.halfcell_degradation_matching_ah(x, *params)

        # Calculate dqdv from full cell profiles
        dq_dv_real = pd.DataFrame(np.gradient(df_real_interped['Q_aligned'], df_real_interped['Voltage_aligned']),
                                  columns=['dQdV']).ewm(0.1).mean()
        dq_dv_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['Q_aligned'], emulated_full_cell_interped['Voltage_aligned']),
            columns=['dQdV']).ewm(0.1).mean()

        # Include original data
        dq_dv_real['Q_aligned'] = df_real_interped['Q_aligned']
        dq_dv_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dq_dv_emulated['Q_aligned'] = emulated_full_cell_interped['Q_aligned']
        dq_dv_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate dQdV and Q over V, aligns real and emulated over V
        voltage_vec = np.linspace(self.FC_LOWER_VOLTAGE, self.FC_UPPER_VOLTAGE, 1001)

        v_dq_dv_interper_real = interp1d(dq_dv_real['Voltage_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                         dq_dv_real['dQdV'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                         bounds_error=False, fill_value=0)
        v_q_interper_real = interp1d(dq_dv_real['Voltage_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                     dq_dv_real['Q_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                     bounds_error=False, fill_value=(0, np.max(df_real_interped['Q_aligned'])))

        v_dq_dv_interper_emulated = interp1d(dq_dv_emulated['Voltage_aligned'].loc[
                                                 ~dq_dv_emulated['Voltage_aligned'].isna()],
                                             dq_dv_emulated['dQdV'].loc[~dq_dv_emulated['Voltage_aligned'].isna()],
                                             bounds_error=False, fill_value=0)
        v_q_interper_emulated = interp1d(dq_dv_emulated['Voltage_aligned'].loc[
                                             ~dq_dv_emulated['Voltage_aligned'].isna()],
                                         dq_dv_emulated['Q_aligned'].loc[~dq_dv_emulated['Voltage_aligned'].isna()],
                                         bounds_error=False, fill_value=(0, np.max(df_real_interped['Q_aligned'])))

        dq_dv_over_v_real = pd.DataFrame(v_dq_dv_interper_real(voltage_vec), columns=['dQdV']).fillna(0)
        dq_dv_over_v_real['Q_aligned'] = v_q_interper_real(voltage_vec)
        dq_dv_over_v_real['Voltage_aligned'] = voltage_vec

        dq_dv_over_v_emulated = pd.DataFrame(v_dq_dv_interper_emulated(voltage_vec), columns=['dQdV']).fillna(0)
        dq_dv_over_v_emulated['Q_aligned'] = v_q_interper_emulated(voltage_vec)
        dq_dv_over_v_emulated['Voltage_aligned'] = voltage_vec

        return (pe_out_zeroed,
                ne_out_zeroed,
                dq_dv_over_v_real,
                dq_dv_over_v_emulated,
                df_real_interped,
                emulated_full_cell_interped)

    def get_dvdq_over_q_from_degradation_matching_ah(self, x, *params):
        """
        This function imposes degradation scaling ,then outputs the dVdQ representation of the emulated cell data.
        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned
        Outputs:
        pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
                offset, and aligned along capacity
        ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
                offset, and aligned along capacity
        dv_dq_real (Dataframe): dVdQ across capacity for the real cell data
        dv_dq_emulated (Dataframe): dVdQ across capacity for the emulated cell data
        df_real_interped (Dataframe): capacity and voltage interpolated evenly across
                capacity for the real cell data
        emulated_full_cell_interped (Dataframe): capacity and voltage interpolated evenly
                across capacity for the emulated cell data
        """

        pe_out_zeroed, ne_out_zeroed, df_real_interped, emulated_full_cell_interped = \
            self.halfcell_degradation_matching_ah(x, *params)

        # Calculate dQdV from full cell profiles
        dv_dq_real = pd.DataFrame(np.gradient(df_real_interped['Voltage_aligned'], df_real_interped['Q_aligned']),
                                  columns=['dVdQ']).ewm(0.1).mean()
        dv_dq_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['Voltage_aligned'], emulated_full_cell_interped['Q_aligned']),
            columns=['dVdQ']).ewm(0.1).mean()

        # Include original data
        dv_dq_real['Q_aligned'] = df_real_interped['Q_aligned']
        dv_dq_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dv_dq_emulated['Q_aligned'] = emulated_full_cell_interped['Q_aligned']
        dv_dq_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Q interpolation not needed, as interpolated over Q by default

        return (pe_out_zeroed,
                ne_out_zeroed,
                dv_dq_real,
                dv_dq_emulated,
                df_real_interped,
                emulated_full_cell_interped)

    def get_v_over_q_from_degradation_matching_ah(self, x, *params):
        """
        This function imposes degradation scaling ,then outputs the V-Q representation of the emulated cell data.
        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned
        Outputs:
        pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
                offset, and aligned along capacity
        ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
                offset, and aligned along capacity
        df_real_interped (Dataframe): capacity and voltage interpolated evenly across
                capacity for the real cell data
        emulated_full_cell_interped (Dataframe): capacity and voltage interpolated evenly
                across capacity for the emulated cell data
        """
        (pe_out_zeroed, ne_out_zeroed, real_aligned, emulated_aligned) = \
            self.halfcell_degradation_matching_ah(x, *params)

        min_soc_full_cell = np.min(real_aligned.loc[~real_aligned.Voltage_aligned.isna()].Q_aligned)
        max_soc_full_cell = np.max(real_aligned.loc[~real_aligned.Voltage_aligned.isna()].Q_aligned)

        soc_vec_full_cell = np.linspace(min_soc_full_cell, max_soc_full_cell, 1001)

        emulated_full_cell_interper = interp1d(
            emulated_aligned.Q_aligned.loc[~real_aligned.Voltage_aligned.isna()],
            emulated_aligned.Voltage_aligned.loc[~real_aligned.Voltage_aligned.isna()],
            bounds_error=False)
        real_full_cell_interper = interp1d(real_aligned.Q_aligned.loc[~real_aligned.Voltage_aligned.isna()],
                                           real_aligned.Voltage_aligned.loc[~real_aligned.Voltage_aligned.isna()],
                                           bounds_error=False)

        # Interpolate the emulated full-cell profile
        emulated_full_cell_interped = pd.DataFrame()
        emulated_full_cell_interped['Q_aligned'] = soc_vec_full_cell
        emulated_full_cell_interped['Voltage_aligned'] = emulated_full_cell_interper(soc_vec_full_cell)

        # Interpolate the true full-cell profile
        df_real_interped = emulated_full_cell_interped.copy()
        df_real_interped['Q_aligned'] = soc_vec_full_cell
        df_real_interped['Voltage_aligned'] = real_full_cell_interper(soc_vec_full_cell)
        return pe_out_zeroed, ne_out_zeroed, df_real_interped, emulated_full_cell_interped

    def get_v_over_q_from_degradation_matching_ah_no_real(self, x, *params):
        """
        This function imposes degradation scaling ,then outputs the V-Q representation of the
        emulated cell data, in the absence of real cell data.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
        pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
                offset, and aligned along capacity
        ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
                offset, and aligned along capacity
        emulated_full_cell_interped (Dataframe): capacity and voltage interpolated evenly
                across capacity for the emulated cell data

        """
        (pe_out_zeroed, ne_out_zeroed, emulated_aligned) = \
            self.halfcell_degradation_matching_ah_no_real(x, *params)

        min_q_full_cell = np.min(emulated_aligned.loc[~emulated_aligned.Voltage_aligned.isna()].Q_aligned)
        max_q_full_cell = np.max(emulated_aligned.loc[~emulated_aligned.Voltage_aligned.isna()].Q_aligned)

        q_vec_full_cell = np.linspace(min_q_full_cell, max_q_full_cell, 1001)

        emulated_full_cell_interper = interp1d(
            emulated_aligned.Q_aligned.loc[~emulated_aligned.Voltage_aligned.isna()],
            emulated_aligned.Voltage_aligned.loc[~emulated_aligned.Voltage_aligned.isna()],
            bounds_error=False)

        # Interpolate the emulated full-cell profile
        emulated_full_cell_interped = pd.DataFrame()
        emulated_full_cell_interped['Q_aligned'] = q_vec_full_cell
        emulated_full_cell_interped['Voltage_aligned'] = emulated_full_cell_interper(q_vec_full_cell)

        return pe_out_zeroed, ne_out_zeroed, emulated_full_cell_interped

    def halfcell_degradation_matching_ah_no_real(self, x, *params):
        """
        Calls underlying functions to impose degradation through electrode
        capacity scale and alignment through LLI. Modifies emulated full cell
        data to be within full cell voltage range and calibrates (zeros) capacity
        at the lowest permissible voltage.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
        pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
                offset, and aligned along capacity
        ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
                offset, and aligned along capacity
        emulated_aligned (Dataframe): full cell data corresponding to the imposed degradation
        """
        lli = x[0]
        q_pe = x[1]
        q_ne = x[2]
        x_ne_2 = x[3]

        pe_pristine, ne_1_pristine, ne_2_pristine_pos, ne_2_pristine_neg = params

        pe_out, ne_out = self._impose_electrode_scale(pe_pristine, ne_1_pristine,
                                                      ne_2_pristine_pos, ne_2_pristine_neg,
                                                      lli, q_pe,
                                                      q_ne,
                                                      x_ne_2)
        # outputs degraded ne and pe (on a AH basis, with electrode alignment (NaNs for voltage, when no overlap))

        emulated_full_cell_with_degradation = pd.DataFrame()
        emulated_full_cell_with_degradation['Q_aligned'] = pe_out['Q_aligned'].copy()
        emulated_full_cell_with_degradation['Voltage_aligned'] = pe_out['Voltage_aligned'] - ne_out['Voltage_aligned']

        # Replace emulated full cell values outside of voltage range with NaN
        emulated_full_cell_with_degradation['Voltage_aligned'].loc[
            emulated_full_cell_with_degradation['Voltage_aligned'] < self.FC_LOWER_VOLTAGE] = np.nan
        emulated_full_cell_with_degradation['Voltage_aligned'].loc[
            emulated_full_cell_with_degradation['Voltage_aligned'] > self.FC_UPPER_VOLTAGE] = np.nan

        # Center the emulated full cell and half cell curves onto the same Q at which the real (degraded)
        # capacity measurement started (self.FC_LOWER_VOLTAGE)
        emulated_full_cell_with_degradation_zeroed = pd.DataFrame()

        emulated_full_cell_with_degradation_zeroed['Voltage_aligned'] = emulated_full_cell_with_degradation[
            'Voltage_aligned']

        zeroing_value = emulated_full_cell_with_degradation['Q_aligned'].loc[
            np.nanargmin(emulated_full_cell_with_degradation['Voltage_aligned'])
        ]

        emulated_full_cell_with_degradation_zeroed['Q_aligned'] = \
            (emulated_full_cell_with_degradation['Q_aligned'] - zeroing_value)

        pe_out_zeroed = pe_out.copy()
        pe_out_zeroed['Q_aligned'] = pe_out['Q_aligned'] - zeroing_value

        ne_out_zeroed = ne_out.copy()
        ne_out_zeroed['Q_aligned'] = ne_out['Q_aligned'] - zeroing_value

        # Interpolate full profiles across same Q range
        min_q = np.min(
            emulated_full_cell_with_degradation_zeroed['Q_aligned'].loc[
                ~emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].isna()])
        max_q = np.max(
            emulated_full_cell_with_degradation_zeroed['Q_aligned'].loc[
                ~emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].isna()])

        emulated_interper = interp1d(emulated_full_cell_with_degradation_zeroed['Q_aligned'].loc[
                                         ~emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].isna()],
                                     emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].loc[
                                         ~emulated_full_cell_with_degradation_zeroed['Voltage_aligned'].isna()],
                                     bounds_error=False)

        q_vec = np.linspace(min_q, max_q, 1001)

        emulated_aligned = pd.DataFrame()
        emulated_aligned['Q_aligned'] = q_vec
        emulated_aligned['Voltage_aligned'] = emulated_interper(q_vec)

        return pe_out_zeroed, ne_out_zeroed, emulated_aligned

    def _get_error_from_degradation_matching_ah(self, x, *params):
        """
        Wrapper function which selects the correct error sub routine and returns its error value.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
            error value (float) - output of the specified error sub function
        """
        error_type = self.error_type
        if error_type == 'V-Q':
            return self._get_error_from_degradation_matching_v_q(x, *params)[0]
        elif error_type == 'dVdQ':
            return self._get_error_from_degradation_matching_dvdq(x, *params)[0]
        elif error_type == 'dQdV':
            return self._get_error_from_degradation_matching_dqdv(x, *params)[0]
        else:
            return self._get_error_from_degradation_matching_v_q(x, *params)[0]

    def _get_error_from_degradation_matching_v_q(self, x, *params):
        """
        Error function returning the mean standardized Euclidean distance of each point of the real curve to
                the closest value on the emulated curve in the V-Q representation.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
            error (float): output of the specified error sub function
            error_vector (array): vector containingEuclidean distance of each point of the real curve to
                the closest value on the emulated curve in the V-Q representation
            xa (Dataframe): real full cell data used for error analysis
            xb (Dataframe): emulated full cell  data used for error analysis
        """

        try:
            (pe_out_zeroed, ne_out_zeroed, real_aligned, emulated_aligned
             ) = self.get_v_over_q_from_degradation_matching_ah(x, *params)

            xa = real_aligned.dropna()
            xb = emulated_aligned.dropna()
            error_matrix = distance.cdist(xa, xb, 'seuclidean')
            error_vector = error_matrix.min(axis=1)
            error = error_vector.mean()
        except ValueError:
            error = 100
            return error, None, None, None
        return error, error_vector, xa, xb

        # Pairwise euclidean from premade dQdV

    def _get_error_from_degradation_matching_dqdv(self, x, *params):
        """
        Error function returning the mean standardized Euclidean distance of each point of the real curve to
                the closest value on the emulated curve in the dQdV representation.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
            error (float): output of the specified error sub function
            error_vector (array): vector containing Euclidean distance of each point of the real curve to
                the closest value on the emulated curve in the dQdV representation
            xa (Dataframe): real full cell data used for error analysis
            xb (Dataframe): emulated full cell  data used for error analysis
        """

        try:
            # Call dQdV generating function
            (pe_out_zeroed,
             ne_out_zeroed,
             dqdv_over_v_real,
             dqdv_over_v_emulated,
             df_real_interped,
             emulated_full_cell_interped) = self.get_dqdv_over_v_from_degradation_matching_ah(x, *params)

            xa = dqdv_over_v_real[['Voltage_aligned', 'dQdV']].dropna()
            xb = dqdv_over_v_emulated[['Voltage_aligned', 'dQdV']].dropna()
            error_matrix = distance.cdist(xa, xb, 'seuclidean')
            error_vector = error_matrix.min(axis=1)
            error = error_vector.mean()

        except ValueError:
            error = 100
            return error, None, None, None
        return error, error_vector, xa, xb

    def _get_error_from_degradation_matching_dvdq(self, x, *params):
        """
        Error function returning the mean standardized Euclidean distance of each point of the real curve to
                the closest value on the emulated curve in the dVdQ representation.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
            error (float): output of the specified error sub function
            error_vector (array): vector containing Euclidean distance of each point of the real curve to
                the closest value on the emulated curve in the dVdQ representation
            xa (Dataframe): real full cell data used for error analysis
            xb (Dataframe): emulated full cell  data used for error analysis
        """

        try:
            (pe_out_zeroed,
             ne_out_zeroed,
             dvdq_over_q_real,
             dvdq_over_q_emulated,
             df_real_interped,
             emulated_full_cell_interped) = self.get_dvdq_over_q_from_degradation_matching_ah(x, *params)

            xa = dvdq_over_q_real[['Q_aligned', 'dVdQ']].dropna()
            xb = dvdq_over_q_emulated[['Q_aligned', 'dVdQ']].dropna()

            # down-select to values with capacity more than 0.5 Ahr to eliminate high-slope region of dVdQ
            xa = xa.loc[(xa.Q_aligned > 0.5)]
            xb = xb.loc[(xb.Q_aligned > 0.5)]

            error_matrix = distance.cdist(xa, xb, 'seuclidean')
            error_vector = error_matrix.min(axis=1)
            error = error_vector.mean()

        except ValueError:
            error = 100
            return error, None, None, None
        return error, error_vector, xa, xb

    def _get_error_from_synthetic_fitting_ah(self, x, *params):
        """
        Wrapper function which selects the correct error sub routine and returns its error value.
        This function is specific to fitting synthetic data rather than real cycling data.

        Inputs:
        x (list): [LLI, q_pe, q_ne, x_ne_2]
        *params:
                pe_pristine (Dataframe): half cell data of the pristine (uncycled) positive
                        electrode
                ne_pristine (Dataframe): half cell data of the pristine (uncycled) negative
                        electrode
                ne_2_pos (Dataframe): half cell data for the positive component of the anode
                ne_2_neg (Dataframe): half cell data for the negative component of the anode
                real_cell_candidate_charge_profile_aligned (Dataframe): columns Q_aligned
                        (evenly spaced) and Voltage_aligned

        Outputs:
            error value (float) - output of the specified error sub function
        """

        error_type = self.error_type

        try:
            if error_type == 'V-Q':
                return self._get_error_from_degradation_matching_v_q(x, *params)[0]
            elif error_type == 'dVdQ':
                return self._get_error_from_degradation_matching_dvdq(x, *params)[0]
            elif error_type == 'dQdV':
                return self._get_error_from_degradation_matching_dvdq(x, *params)[0]
            else:
                return self._get_error_from_degradation_matching_v_q(x, *params)[0]
        except RuntimeError:
            print("Can't return error")
            return 100

    def intracell_values_wrapper_ah(self,
                                    cycle_index,
                                    cell_struct,
                                    degradation_bounds=None
                                    ):
        """
        Wrapper function to solve capacity sizing and offset of reference electrodes to real full cell cycle data.

        Inputs:
        cycle_index (int): the index of the cycle of interest of the structured real cycling data
        cell_struct (MaccorDatapath): BEEP structured cycling data

        Outputs:
        loss_dict (dict): dictionary with key of cycle index and entry of a list of
                error, lli_opt, q_pe_opt, q_ne_opt, x_ne_2, Q_li
        profiles_dict (dict): dictionary with key of cycle index and entry of a dictionary
                containing various key/entry pairs of resulting from the fitting
        """
        if degradation_bounds is None:
            degradation_bounds = ((0, 3),  # LLI
                                  (2.5, 6.5),  # q_pe
                                  (2.5, 6.5),  # q_ne
                                  (1, 1),  # (-1,1) x_ne_2
                                  )

        real_cell_candidate_charge_profile_aligned = self.process_beep_cycle_data_for_candidate_halfcell_analysis_ah(
            cell_struct,
            cycle_index)

        degradation_optimization_result = differential_evolution(self._get_error_from_degradation_matching_ah,
                                                                 degradation_bounds,
                                                                 args=(self.pe_pristine,
                                                                       self.ne_1_pristine,
                                                                       self.ne_2_pristine_pos,
                                                                       self.ne_2_pristine_neg,
                                                                       real_cell_candidate_charge_profile_aligned
                                                                       ),
                                                                 strategy='best1bin', maxiter=100000,
                                                                 popsize=15, tol=0.001, mutation=0.5,
                                                                 recombination=0.7,
                                                                 seed=1,
                                                                 callback=None, disp=False, polish=True,
                                                                 init='latinhypercube',
                                                                 atol=0, updating='deferred', workers=-1,
                                                                 constraints=()
                                                                 )
        #         print(degradation_optimization_result.x) #BVV

        (pe_out_zeroed,
         ne_out_zeroed,
         dqdv_over_v_real,
         dqdv_over_v_emulated,
         df_real_interped,
         emulated_full_cell_interped) = self.get_dqdv_over_v_from_degradation_matching_ah(
            degradation_optimization_result.x,
            self.pe_pristine,
            self.ne_1_pristine,
            self.ne_2_pristine_pos,
            self.ne_2_pristine_neg,
            real_cell_candidate_charge_profile_aligned)
        #
        electrode_info_df = get_electrode_info_ah(pe_out_zeroed, ne_out_zeroed)
        #
        error = degradation_optimization_result.fun
        lli_opt = degradation_optimization_result.x[0]
        q_pe_opt = degradation_optimization_result.x[1]
        q_ne_opt = degradation_optimization_result.x[2]
        x_ne_2 = degradation_optimization_result.x[3]

        loss_dict = {cycle_index: np.append([error, lli_opt, q_pe_opt, q_ne_opt,
                                             x_ne_2],
                                            electrode_info_df.iloc[-1].values)
                     }

        profiles_per_cycle_dict = {
           'NE_zeroed': ne_out_zeroed,
           'PE_zeroed': pe_out_zeroed,
           'dQdV_over_v_real': dqdv_over_v_real,
           'dQdV_over_v_emulated': dqdv_over_v_emulated,
           'df_real_interped': df_real_interped,
           'emulated_full_cell_interped': emulated_full_cell_interped,
           'real_cell_candidate_charge_profile_aligned': real_cell_candidate_charge_profile_aligned
                                   }

        profiles_dict = {cycle_index: profiles_per_cycle_dict}

        return loss_dict, profiles_dict

    def solve_emulated_degradation(self,
                                   forward_simulated_profile,
                                   degradation_bounds=None
                                   ):

        """


        """

        if degradation_bounds is None:
            degradation_bounds = ((0, 3),  # LLI
                                  (2.5, 6.5),  # q_pe
                                  (2.5, 6.5),  # q_ne
                                  (1, 1),  # (-1,1) x_ne_2
                                  )

        degradation_optimization_result = differential_evolution(self._get_error_from_synthetic_fitting_ah,
                                                                 degradation_bounds,
                                                                 args=(self.pe_pristine,
                                                                       self.ne_1_pristine,
                                                                       self.ne_2_pristine_pos,
                                                                       self.ne_2_pristine_neg,
                                                                       forward_simulated_profile,
                                                                       ),
                                                                 strategy='best1bin', maxiter=100000,
                                                                 popsize=15, tol=0.001, mutation=0.5,
                                                                 recombination=0.7,
                                                                 seed=1,
                                                                 callback=None, disp=False, polish=True,
                                                                 init='latinhypercube',
                                                                 atol=0, updating='deferred', workers=-1,
                                                                 constraints=()
                                                                 )

        return degradation_optimization_result


# TODO revisit this function
def blend_electrodes(electrode_1, electrode_2_pos, electrode_2_neg, x_2):
    """
    Blends two electrode materials from their SOC-V profiles to form a blended electrode.

    Inputs:
    electrode_1: Primary material in electrode, typically Gr. DataFrame supplied with SOC evenly spaced and voltage.
    electrode_2: Secondary material in electrode, typically Si. DataFrame supplied with SOC evenly spaced and
        voltage as an additional column.
    x_2: Fraction of electrode_2 material's capacity (not mass). Supplied as scalar value.

    Outputs:
    df_blended_soc_mod (Dataframe): blended electrode with SOC_aligned and Voltage_aligned columns
    """
    if electrode_2_pos.empty:
        df_blended = electrode_1
        return df_blended

    if electrode_2_neg.empty:
        electrode_2 = electrode_2_pos
        x_2 = np.abs(x_2)
    elif x_2 > 0:
        electrode_2 = electrode_2_pos
    else:
        electrode_2 = electrode_2_neg
        x_2 = np.abs(x_2)

    electrode_1_interper = interp1d(electrode_1['Voltage_aligned'], electrode_1['SOC_aligned'], bounds_error=False,
                                    fill_value='extrapolate')
    electrode_2_interper = interp1d(electrode_2['Voltage_aligned'], electrode_2['SOC_aligned'], bounds_error=False,
                                    fill_value='extrapolate')

    voltage_vec = np.linspace(np.min((np.min(electrode_1['Voltage_aligned']),
                                      np.min(electrode_2['Voltage_aligned']))),
                              np.max((np.max(electrode_1['Voltage_aligned']),
                                      np.max(electrode_2['Voltage_aligned']))),
                              1001)
    electrode_1_voltage_aligned = pd.DataFrame(electrode_1_interper(voltage_vec), columns=['SOC'])
    electrode_2_voltage_aligned = pd.DataFrame(electrode_2_interper(voltage_vec), columns=['SOC'])
    electrode_1_voltage_aligned['Voltage'] = voltage_vec
    electrode_2_voltage_aligned['Voltage'] = voltage_vec

    df_blend_voltage_aligned = pd.DataFrame(
        (1 - x_2) * electrode_1_voltage_aligned['SOC'] + x_2 * electrode_2_voltage_aligned['SOC'], columns=['SOC'])
    df_blend_voltage_aligned['Voltage'] = electrode_1_voltage_aligned.merge(electrode_2_voltage_aligned,
                                                                            on='Voltage')['Voltage']

    df_blended_interper = interp1d(df_blend_voltage_aligned['SOC'], df_blend_voltage_aligned['Voltage'],
                                   bounds_error=False)
    soc_vec = np.linspace(0, 100, 1001)

    df_blended = pd.DataFrame(df_blended_interper(soc_vec), columns=['Voltage_aligned'])
    df_blended['SOC_aligned'] = soc_vec

    # Modify NE to fully span 100% SOC within its valid voltage window
    df_blended_soc_mod_interper = interp1d(df_blended['SOC_aligned'].loc[~df_blended['Voltage_aligned'].isna()],
                                           df_blended['Voltage_aligned'].loc[~df_blended['Voltage_aligned'].isna()],
                                           bounds_error=False)
    soc_vec = np.linspace(np.min(df_blended['SOC_aligned'].loc[~df_blended['Voltage_aligned'].isna()]),
                          np.max(df_blended['SOC_aligned'].loc[~df_blended['Voltage_aligned'].isna()]),
                          1001)
    df_blended_soc_mod = pd.DataFrame(df_blended_soc_mod_interper(soc_vec), columns=['Voltage_aligned'])
    df_blended_soc_mod['SOC_aligned'] = soc_vec / np.max(soc_vec) * 100
    return df_blended_soc_mod


def get_electrode_info_ah(pe_out_zeroed, ne_out_zeroed):
    """
    Calculates a variety of half-cell metrics at various positions in the full-cell profile.

    Inputs:
    pe_out_zeroed (Dataframe): cathode capacity and voltage columns scaled,
            offset, and aligned along capacity
    ne_out_zeroed (Dataframe): anode capacity and voltage columns scaled,
            offset, and aligned along capacity

    Outputs:
    electrode_info_df (Dataframe): dataframe containing a variety of half-cell metrics
        at various positions in the emulated full-cell profile.

        pe_voltage_FC4p2V: voltage of the positive electrode (catahode) corresponding
            to the full cell at 4.2V
        ...
        pe_voltage_FC2p7V: voltage of the positive electrode (catahode) corresponding
            to the full cell at 2.7V

        pe_soc_FC4p2V: state of charge of the positive electrode corresponding
            to the full cell at 4.2V
        ...
        pe_soc_FC2p7V: state of charge of the positive electrode corresponding
            to the full cell at 2.7V

        ne_voltage_FC4p2V: voltage of the negative electrode (anode) corresponding
            to the full cell at 4.2V
        ...
        ne_voltage_FC2p7V: voltage of the negative electrode (anode) corresponding
            to the full cell at 2.7V

        ne_soc_FC4p2V: state of charge of the anode electrode corresponding
            to the full cell at 4.2V
        ...
        ne_soc_FC2p7V: state of charge of the anode electrode corresponding
            to the full cell at 2.7V

        Q_fc: capacity of the full cecll within the full cell voltage limits
        q_pe: capacity of the cathode
        q_ne: capacity of the anode [Ahr]
        Q_li
    """
    pe_minus_ne_zeroed = pd.DataFrame(pe_out_zeroed['Voltage_aligned'] - ne_out_zeroed['Voltage_aligned'],
                                      columns=['Voltage_aligned'])
    pe_minus_ne_zeroed['Q_aligned'] = pe_out_zeroed['Q_aligned']

    electrode_info_df = pd.DataFrame(index=[0])

    electrode_info_df['pe_voltage_FC4p2V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.2))].Voltage_aligned
    electrode_info_df['pe_voltage_FC4p1V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.1))].Voltage_aligned
    electrode_info_df['pe_voltage_FC4p0V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.0))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p9V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.9))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p8V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.8))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p7V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.7))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p6V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.6))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p5V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.5))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p4V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.4))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p3V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.3))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p2V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.2))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p1V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.1))].Voltage_aligned
    electrode_info_df['pe_voltage_FC3p0V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.0))].Voltage_aligned
    electrode_info_df['pe_voltage_FC2p9V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.9))].Voltage_aligned
    electrode_info_df['pe_voltage_FC2p8V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.8))].Voltage_aligned
    electrode_info_df['pe_voltage_FC2p7V'] = pe_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.7))].Voltage_aligned

    electrode_info_df['pe_soc_FC4p2V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.2))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 4.2V
    electrode_info_df['pe_soc_FC4p1V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.1))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 4.1V
    electrode_info_df['pe_soc_FC4p0V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.0))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 4.0V
    electrode_info_df['pe_soc_FC3p9V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.9))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.9V
    electrode_info_df['pe_soc_FC3p8V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.8))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.8V
    electrode_info_df['pe_soc_FC3p7V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.7))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.7V
    electrode_info_df['pe_soc_FC3p6V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.6))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.6V
    electrode_info_df['pe_soc_FC3p5V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.5))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.5V
    electrode_info_df['pe_soc_FC3p4V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.4))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.4V
    electrode_info_df['pe_soc_FC3p3V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.3))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.3V
    electrode_info_df['pe_soc_FC3p2V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.2))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.2V
    electrode_info_df['pe_soc_FC3p1V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.1))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.1V
    electrode_info_df['pe_soc_FC3p0V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.0))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.0V
    electrode_info_df['pe_soc_FC2p9V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.9))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 2.9V
    electrode_info_df['pe_soc_FC2p8V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.8))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 2.8V
    electrode_info_df['pe_soc_FC2p7V'] = (
                (pe_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.7))].Q_aligned -
                 np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]))
                )  # 2.7V

    electrode_info_df['ne_voltage_FC4p2V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.2))].Voltage_aligned
    electrode_info_df['ne_voltage_FC4p1V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.1))].Voltage_aligned
    electrode_info_df['ne_voltage_FC4p0V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.0))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p9V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.9))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p8V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.8))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p7V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.7))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p6V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.6))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p5V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.5))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p4V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.4))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p3V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.3))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p2V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.2))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p1V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.1))].Voltage_aligned
    electrode_info_df['ne_voltage_FC3p0V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.0))].Voltage_aligned
    electrode_info_df['ne_voltage_FC2p9V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.9))].Voltage_aligned
    electrode_info_df['ne_voltage_FC2p8V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.8))].Voltage_aligned
    electrode_info_df['ne_voltage_FC2p7V'] = ne_out_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.7))].Voltage_aligned

    electrode_info_df['ne_soc_FC4p2V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.2))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 4.2V
    electrode_info_df['ne_soc_FC4p1V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.1))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 4.1V
    electrode_info_df['ne_soc_FC4p0V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.0))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 4.0V
    electrode_info_df['ne_soc_FC3p9V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.9))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.9V
    electrode_info_df['ne_soc_FC3p8V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.8))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3Q_aligned.8V
    electrode_info_df['ne_soc_FC3p7V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.7))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.7V
    electrode_info_df['ne_soc_FC3p6V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.6))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.6V
    electrode_info_df['ne_soc_FC3p5V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.5))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.5V
    electrode_info_df['ne_soc_FC3p4V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.4))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.4V
    electrode_info_df['ne_soc_FC3p3V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.3))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.3V
    electrode_info_df['ne_soc_FC3p2V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.2))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.2V
    electrode_info_df['ne_soc_FC3p1V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.1))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.1V
    electrode_info_df['ne_soc_FC3p0V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 3.0))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 3.0V
    electrode_info_df['ne_soc_FC2p9V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.9))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 2.9V
    electrode_info_df['ne_soc_FC2p8V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.8))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 2.8V
    electrode_info_df['ne_soc_FC2p7V'] = (
                (ne_out_zeroed.loc[np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 2.7))].Q_aligned -
                 np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])) / (
                        np.max(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) -
                        np.min(ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]))
                )  # 2.7V

    electrode_info_df['Q_fc'] = pe_minus_ne_zeroed.loc[
        np.argmin(np.abs(pe_minus_ne_zeroed.Voltage_aligned - 4.20))].Q_aligned

    electrode_info_df['Q_pe'] = np.max(
        pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()]) - np.min(
        pe_out_zeroed['Q_aligned'].loc[~pe_out_zeroed['Voltage_aligned'].isna()])

    electrode_info_df['Q_ne'] = np.max(
        ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()]) - np.min(
        ne_out_zeroed['Q_aligned'].loc[~ne_out_zeroed['Voltage_aligned'].isna()])

    electrode_info_df['Q_li'] = np.max(
        pe_minus_ne_zeroed['Q_aligned'].loc[~pe_minus_ne_zeroed.Voltage_aligned.isna()]) - np.min(
        pe_minus_ne_zeroed['Q_aligned'].loc[~pe_minus_ne_zeroed.Voltage_aligned.isna()])

    return electrode_info_df



"""Classes and functions for handling BioLogic battery cyclers.
"""

import hashlib
import os.path
import json
from datetime import datetime
import pytz

import pandas as pd

from beep.structure.base import BEEPDatapath
from beep.conversion_schemas import BIOLOGIC_CONFIG


class BiologicDatapath(BEEPDatapath):
    """Datapath for ingesting and structuring BioLogic cycler data.
    """

    @staticmethod
    def _get_file_type(path, search_lines=200):
        """
        Checks on the file and gets the right separators and header length.
        Should enable seemless processing of either manual exported file (.txt) or
        automatically exported file (.csv)

        Args:
            path (str): file path to data file
            search_lines (int): number of lines to read when looking for header line

        Returns:
            sep (str): separation character between fields
            encoding (str): file encoding to expect based on extension
            header_starts_line (int): line number of the header line
            data_starts_line (int): line number where data values start
        """
        if os.path.splitext(path)[1] == ".csv":
            sep = ";"
            encoding = "utf-8"
        elif os.path.splitext(path)[1] == ".txt":
            sep = "\t"
            encoding = "iso-8859-1"
        else:
            raise TypeError("Unable to determine separator character from file extension")

        with open(path, "rb") as f:
            i = 1
            header_starts_line = None
            while header_starts_line is None:
                line = f.readline()
                if b'Ecell/V' in line and b'Variables' not in line:
                    header_starts_line = i
                    data_starts_line = i + 1
                i += 1
                if i > search_lines:
                    raise LookupError("Unable to find the header line in first {} lines of file".format(search_lines))

        return sep, encoding, header_starts_line, data_starts_line

    @classmethod
    def from_file(cls, path, mapping_file=None):
        """Creates a BEEPDatapath from a raw BioLogic battery cycler output file.

        Args:
            path (str): file path to data file
            mapping_file (str): file path to mapping file containing step transitions where cycle_index should
                increment

        Returns:
            BiologicDatapath
        """

        sep, encoding, header_line, data_starts_line = cls._get_file_type(path)
        column_map = BIOLOGIC_CONFIG["data_columns"]

        raw = dict()
        i = 0
        with open(path, "rb") as f:

            empty_lines = 0  # used to find the end of the data entries.
            while empty_lines < 2:  # finding 2 empty lines in a row => EOF
                line = f.readline()
                i += 1
                if i == header_line:
                    header = str(line.decode(encoding=encoding))
                    columns = header.split(sep)
                    for c in columns:
                        raw[c] = list()
                if i >= data_starts_line:
                    line = line.decode(encoding=encoding)
                    if len(line) == 0:
                        empty_lines += 1
                        continue
                    items = line.split(sep)
                    for ci in range(len(items)):
                        column_name = columns[ci]
                        data_type = column_map.get(column_name, dict()).get(
                            "data_type", str
                        )
                        scale = column_map.get(column_name, dict()).get("scale", 1.0)
                        item = items[ci]
                        if data_type == "int":
                            item = int(float(item))
                        if data_type == "float":
                            item = float(item) * scale
                        raw[column_name].append(item)

        if "cycle_index" not in columns and not mapping_file:
            raw["cycle_index"] = [int(float(i)) for i in raw["cycle number"]]
        elif "cycle_index" not in columns and mapping_file:
            if "Loop" in raw.keys():
                raw["cycle_index"] = get_cycle_index(raw["Ns"], mapping_file, loop_list=raw["Loop"])
            else:
                raw["cycle_index"] = get_cycle_index(raw["Ns"], mapping_file)

        data = dict()
        for column_name in column_map.keys():
            data[column_map[column_name]["beep_name"]] = raw[column_name]
        data["data_point"] = list(range(1, len(raw["cycle number"]) + 1))

        data = pd.DataFrame(data)
        metadata_path = path.replace(".csv", ".mpl")
        metadata = cls.parse_metadata(metadata_path)
        metadata["filename"] = path
        metadata["_today_datetime"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # standardizing time format
        pacific = pytz.timezone("US/Pacific")
        utc = pytz.timezone("UTC")
        data["date_time_iso"] = data["date_time"].apply(
            lambda x: pacific.localize(datetime.strptime(x, "%m/%d/%Y %H:%M:%S.%f"),
                                       is_dst=True).astimezone(utc).isoformat()
        )
        data["test_time"] = data["date_time"].apply(
            lambda x: (datetime.strptime(x, "%m/%d/%Y %H:%M:%S.%f") -
                       datetime.strptime(data["date_time"].iloc[0], "%m/%d/%Y %H:%M:%S.%f")).total_seconds()
        )
        paths = {
            "raw": path,
            "metadata": metadata_path
        }

        return cls(data, metadata, paths)

    @staticmethod
    def parse_metadata(metadata_path):
        """Extracts BioLogic metadata from metadata file.

        Iterates through a biologic metadata file and extracts the meta fields:
            - cell_id
            - barcode
            - protocol

        Args:
            metadata_path (str): path to metadata file

        Returns:
            (dict): dictionary with metadata fields
        """

        flag_cell_id = False
        flag_barcode = False
        flag_protocol = False
        flag_protocol_start = False
        protocol_text = ""
        max_lines = 10000

        metadata = dict()

        with open(metadata_path, "rb") as f:

            i = 0
            while True:
                line = f.readline()
                line = str(line)
                if line.startswith("b'"):
                    line = line[2:]
                if line.endswith("'"):
                    line = line[:-1]
                if line.endswith("\\r\\n"):
                    line = line[:-4]

                if line.strip().split(" : ")[0] == "Run on channel":
                    channel_id = line.strip().split(' : ')[1]
                    channel_id = channel_id.split(' ')[0]
                    channel_id = int(str(ord(channel_id[0]) - 64) + channel_id[1:])
                    metadata["channel_id"] = channel_id
                    flag_cell_id = True

                if "Comments" in line.strip().split(" : ")[0]:
                    barcode = line.strip().split(" : ")[-1]
                    metadata["barcode"] = barcode
                    flag_barcode = True

                if "Cycle Definition" in line.strip().split(" : ")[0]:
                    protocol_text += line + "\n"
                    flag_protocol_start = True

                if flag_protocol_start:
                    if line == "":
                        flag_protocol = True
                        protocol = hashlib.md5(protocol_text.encode()).digest()
                        metadata["protocol"] = protocol
                    else:
                        protocol_text += line + "\n"

                done = flag_cell_id and flag_barcode and flag_protocol
                if done:
                    break

                i += 1
                if i > max_lines:
                    break
        return metadata


def get_cycle_index(ns_list, serialized_transition_fp, loop_list=None):
    """
    Processes CSV files generated from several biologic techniques
    and creates a new set of CSVs with an additional "cycle_index" column.

    Args:
        df (pandas.DataFrame): data frame of biologic file
        serialized_transition_fp (path): path to mapping file containing step transitions where
            cycle index should increment

    Returns:
        df (pandas.DataFrame): data frame of biologic file with cycle_index and Tech Num added

    """

    serializer = CycleTransitionRulesSerializer()
    cycle_num = 1

    with open(serialized_transition_fp, "r") as f:
        data = f.read()
        cycle_transition_rules = serializer.parse_json(data)

    cycle_num += cycle_transition_rules.adv_cycle_on_start

    prev_seq_num = int(ns_list[0])
    if loop_list:
        prev_loop_num = int(loop_list[0])
    cycle_nums = []
    tech_nums = []
    # TODO speed up by reducing logic and use list comprehension
    if loop_list:
        for indx, ns in enumerate(ns_list):
            seq_num = int(ns)
            loop_num = int(loop_list[indx])
            # a transition may occur because of a loop technique or a loop seq,
            # it is possible to double count cycle advances if we don't handle them separately
            if loop_num != prev_loop_num:
                cycle_num += cycle_transition_rules.adv_cycle_on_tech_loop
            elif seq_num != prev_seq_num:
                transition = (prev_seq_num, seq_num)
                cycle_num += cycle_transition_rules.adv_cycle_seq_transitions.get(
                    transition, 0
                )
            prev_loop_num = loop_num
            prev_seq_num = seq_num
            cycle_nums.append(cycle_num)
            tech_nums.append(cycle_transition_rules.tech_num)
    else:
        for indx, ns in enumerate(ns_list):
            seq_num = int(ns)
            if seq_num != prev_seq_num:
                transition = (prev_seq_num, seq_num)
                cycle_num += cycle_transition_rules.adv_cycle_seq_transitions.get(
                    transition, 0
                )
            prev_seq_num = seq_num
            cycle_nums.append(cycle_num)
            tech_nums.append(cycle_transition_rules.tech_num)

    return cycle_nums


class CycleTransitionRules:
    def __init__(
        self,
        tech_num,
        tech_does_loop,
        adv_cycle_on_start,
        adv_cycle_on_tech_loop,
        adv_cycle_seq_transitions,
        debug_adv_cycle_on_step_transitions={},
    ):
        self.tech_num = tech_num
        self.tech_does_loop = tech_does_loop
        self.adv_cycle_on_start = adv_cycle_on_start
        self.adv_cycle_on_tech_loop = adv_cycle_on_tech_loop
        self.adv_cycle_seq_transitions = adv_cycle_seq_transitions
        self.debug_adv_cycle_on_step_transitions = debug_adv_cycle_on_step_transitions

    def __repr__(self):
        return (
            "{\n"
            + "  tech_num: {},\n".format(self.tech_num)
            + "  tech_does_loop: {},\n".format(self.tech_does_loop)
            + "  adv_cycle_on_start: {},\n".format(self.adv_cycle_on_start)
            + "  adv_cycle_on_tech_loop: {},\n".format(self.adv_cycle_on_tech_loop)
            + "  adv_cycle_seq_transitions: {},\n".format(
                self.adv_cycle_seq_transitions
            )
            + "  debug_adv_cycle_on_step_transitions: {},\n".format(
                self.debug_adv_cycle_on_step_transitions
            )
            + "}\n"
        )


class CycleTransitionRulesSerializer:
    def json(self, cycle_transition_rules, indent=2):
        parseable_adv_cycle_seq_transitions = []
        for (
            s,
            t,
        ), adv_cycle_count in cycle_transition_rules.adv_cycle_seq_transitions.items():
            parseable_adv_cycle_seq_transitions.append(
                {
                    "source": s,
                    "target": t,
                    "adv_cycle_count": adv_cycle_count,
                }
            )

        parseable_debug_adv_cycle_on_step_transitions = []
        for (
            s,
            t,
        ), adv_cycle_count in (
            cycle_transition_rules.debug_adv_cycle_on_step_transitions.items()
        ):
            parseable_debug_adv_cycle_on_step_transitions.append(
                {
                    "source": s,
                    "target": t,
                    "adv_cycle_count": adv_cycle_count,
                }
            )

        obj = {
            "tech_num": cycle_transition_rules.tech_num,
            "tech_does_loop": cycle_transition_rules.tech_does_loop,
            "adv_cycle_on_start": cycle_transition_rules.adv_cycle_on_start,
            "adv_cycle_on_tech_loop": cycle_transition_rules.adv_cycle_on_tech_loop,
            # these are (int, int) -> int maps, tuples cannot be represented in json
            "adv_cycle_seq_transitions": parseable_adv_cycle_seq_transitions,
            "debug_adv_cycle_on_step_transitions": parseable_debug_adv_cycle_on_step_transitions,
        }

        return json.dumps(obj, indent=indent)

    def parse_json(self, serialized):
        data = json.loads(serialized)

        tech_num = data["tech_num"]
        tech_does_loop = data["tech_does_loop"]
        adv_cycle_on_start = data["adv_cycle_on_start"]
        adv_cycle_on_tech_loop = data["adv_cycle_on_tech_loop"]

        adv_cycle_seq_transitions = {}
        for d in data["adv_cycle_seq_transitions"]:
            transition = (d["source"], d["target"])
            adv_cycle_seq_transitions[transition] = d["adv_cycle_count"]

        debug_adv_cycle_on_step_transitions = {}
        for d in data["debug_adv_cycle_on_step_transitions"]:
            transition = (d["source"], d["target"])
            debug_adv_cycle_on_step_transitions[transition] = d["adv_cycle_count"]

        return CycleTransitionRules(
            tech_num,
            tech_does_loop,
            adv_cycle_on_start,
            adv_cycle_on_tech_loop,
            adv_cycle_seq_transitions,
            debug_adv_cycle_on_step_transitions,
        )



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

For creating features and organizing them into datasets.

"""
import os
import copy
import abc
import json
import hashlib
from typing import Union, Tuple, List

import pandas as pd
from monty.io import zopen
from monty.json import MSONable, MontyDecoder
from monty.serialization import loadfn, dumpfn

from beep.structure.base import BEEPDatapath


class BEEPFeaturizationError(BaseException):
    """Raise when a featurization-specific error occurs"""
    pass


class BEEPFeatureMatrixError(BaseException):
    """ Raise when there is a BEEP-specific problem with a dataset"""
    pass


class BEEPFeaturizer(MSONable, abc.ABC):
    """
    Base class for all beep feature generation.

    From a structured battery file representing many cycles of one cell,
    (AKA a structured datapath), produce a feature vector.

    Works for generating both
     - Vectors X to use as training vectors
     - Vectors or scalars y to use as ML targets
        (as problems may have multiple metrics to predict)

    """

    DEFAULT_HYPERPARAMETERS = {}

    def __init__(self, structured_datapath: Union[BEEPDatapath, None], hyperparameters: Union[dict, None] = None):
        # If all required hyperparameters are specified, use those
        # If some subset of required hyperparameters are specified, throw error
        # If no hyperparameters are specified, use defaults
        if hyperparameters:
            if all(k in hyperparameters for k in self.DEFAULT_HYPERPARAMETERS):
                self.hyperparameters = hyperparameters
            else:
                raise BEEPFeaturizationError(
                    f"Features cannot be created with incomplete set of "
                    f"hyperparameters {hyperparameters.keys()} < "
                    f"{self.DEFAULT_HYPERPARAMETERS.keys()}!")
        else:
            self.hyperparameters = self.DEFAULT_HYPERPARAMETERS

        if not (structured_datapath is None or structured_datapath.is_structured):
            raise BEEPFeaturizationError("BEEPDatapath input is not structured!")
        self.datapath = structured_datapath

        self.features = None

        # In case these features are loaded from file
        # Allow attrs which can hold relevant metadata without having
        # to reload the original datapath
        self.paths = self.datapath.paths if self.datapath else {}
        self.metadata = self.datapath.metadata.raw if self.datapath else {}
        self.linked_semiunique_id = self.datapath.semiunique_id if self.datapath else None

    @abc.abstractmethod
    def validate(self) -> Tuple[bool, Union[str, None]]:
        """
        Validate a featurizer on it's ingested datapath.

        Returns:
            (bool, str/None): The validation result and it's message.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_features(self) -> None:
        """
        Should assign a dataframe to self.features.

        Returns:
            None
        """
        raise NotImplementedError

    def as_dict(self):
        """Serialize a BEEPDatapath as a dictionary.

        Must not be loaded from legacy.

        Returns:
            (dict): corresponding to dictionary for serialization.

        """

        if self.features is None:
            raise BEEPFeaturizationError("Cannot serialize features which have not been generated.")

        features = self.features.to_dict("list")

        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,

            # Core parts of BEEPFeaturizer
            "features": features,
            "hyperparameters": self.hyperparameters,
            "paths": self.paths,
            "metadata": self.metadata,
            "linked_datapath_semiunique_id": self.linked_semiunique_id
        }

    @classmethod
    def from_dict(cls, d):
        """Create a BEEPDatapath object from a dictionary.

        Args:
            d (dict): dictionary represenation.

        Returns:
            beep.structure.ProcessedCyclerRun: deserialized ProcessedCyclerRun.
        """

        # no need for original datapath
        bf = cls(structured_datapath=None, hyperparameters=d["hyperparameters"])
        bf.features = pd.DataFrame(d["features"])
        bf.paths = d["paths"]
        bf.metadata = d["metadata"]
        bf.linked_semiunique_id = d["linked_datapath_semiunique_id"]
        return bf

    @classmethod
    def from_json_file(cls, filename):
        """Load a structured run previously saved to file.

        .json.gz files are supported.

        Loads a BEEPFeaturizer from json.

        Can be used in combination with files serialized with BEEPFeatures.to_json_file.

        Args:
            filename (str, Pathlike): a json file from a structured run, serialzed with to_json_file.

        Returns:
            None
        """
        with zopen(filename, "r") as f:
            d = json.load(f)

        # Add this structured file path to the paths dict
        paths = d.get("paths", {})
        paths["features"] = os.path.abspath(filename)
        d["paths"] = paths
        return cls.from_dict(d)

    def to_json_file(self, filename):
        """Save a BEEPFeatures to disk as a json.

        .json.gz files are supported.

        Not named from_json to avoid conflict with MSONable.from_json(*)

        Args:
            filename (str, Pathlike): The filename to save the file to.
            omit_raw (bool): If True, saves only structured (NOT RAW) data.
                More efficient for saving/writing to disk.

        Returns:
            None
        """
        d = self.as_dict()
        dumpfn(d, filename)


class BEEPFeatureMatrix(MSONable):
    """
    Create an (n battery cycler files) x (k features) array composed of
    m BEEPFeaturizer objects.

    Args:
        beepfeaturizers ([BEEPFeaturizer]): A list of BEEPFeaturizer objects

    """

    OP_DELIMITER = "::"

    def __init__(self, beepfeaturizers: List[BEEPFeaturizer]):

        if beepfeaturizers:
            dfs_by_file = {bf.paths.get("structured", "no file found"): [] for bf in beepfeaturizers}
            # big_df_rows = {bf.__class__.__name__: [] for bf in beepfeaturizers}
            unique_features = {}
            for i, bf in enumerate(beepfeaturizers):
                if bf.features is None:
                    raise BEEPFeatureMatrixError(f"BEEPFeaturizer {bf} has not created features")
                elif bf.features.shape[0] != 1:
                    raise BEEPFeatureMatrixError(f"BEEPFeaturizer {bf} features are not 1-dimensional.")
                else:
                    bfcn = bf.__class__.__name__

                    fname = bf.paths.get("structured", None)
                    if not fname:
                        raise BEEPFeatureMatrixError(
                            "Cannot join features automatically as no linking can be done "
                            "based on original structured filename."
                        )

                    # Check for any possible feature collisions using identical featurizers
                    # on identical files

                    # sort params for this featurizer obj by key
                    params = sorted(list(bf.hyperparameters.items()), key=lambda x: x[0])

                    # Prevent identical features from identical input files
                    # create a unique operation string for the application of this featurizer
                    # on a specific file, this op string will be the same as long as
                    # the featurizer class name, hyperparameters, and class are the same

                    param_str = "-".join([f"{k}:{v}" for k, v in params])
                    param_hash = hashlib.sha256(param_str.encode("utf-8")).hexdigest()

                    # Get an id for this featurizer operation (including hyperparameters)
                    # regardless of the file it is applied on
                    feature_op_id = f"{bfcn}{self.OP_DELIMITER}{param_hash}"

                    # Get an id for this featurizer operation (including hyperparameters)
                    # on THIS SPECIFIC file.
                    file_feature_op_id = f"{fname}{self.OP_DELIMITER}{bfcn}{self.OP_DELIMITER}{param_hash}"

                    # Get a unique id for every feature generated by a specific
                    # featurizer on a specific file.
                    this_file_feature_columns_ids = \
                        [
                            f"{file_feature_op_id}{self.OP_DELIMITER}{c}" for c in bf.features.columns
                        ]

                    # Check to make sure there are no duplicates of the exact same feature for
                    # the exact same featurizer with the exact same hyperparameters on the exact
                    # same file.
                    collisions = {c: f for c, f in unique_features.items() if c in this_file_feature_columns_ids}
                    if collisions:
                        raise BEEPFeatureMatrixError(
                            f"Multiple features generated with identical classes and identical hyperparameters"
                            f" attempted to be joined into same dataset; \n"
                            f"{bfcn} features collide with existing: \n{collisions}"
                        )
                    for c in this_file_feature_columns_ids:
                        unique_features[c] = bfcn

                    # Create consistent scheme for naming features regardless of file
                    df = copy.deepcopy(bf.features)
                    consistent_column_names = [f"{c}{self.OP_DELIMITER}{feature_op_id}" for c in df.columns]
                    df.columns = consistent_column_names

                    df.index = [fname] * df.shape[0]
                    df.index.rename("filename", inplace=True)
                    dfs_by_file[fname].append(df)

            rows = []
            for filename, dfs in dfs_by_file.items():
                row = pd.concat(dfs, axis=1)
                row = row[sorted(row.columns)]
                rows.append(row)
            self.matrix = pd.concat(rows, axis=0)

        else:
            self.matrix = None

        self.featurizers = beepfeaturizers

    def as_dict(self):
        """Serialize a BEEPDatapath as a dictionary.

        Must not be loaded from legacy.

        Returns:
            (dict): corresponding to dictionary for serialization.

        """

        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,

            # Core parts of BEEPFeaturizer
            "featurizers": [f.as_dict() for f in self.featurizers],
            "matrix": self.matrix.to_dict("list"),
        }

    @classmethod
    def from_dict(cls, d):
        """Create a BEEPDatapath object from a dictionary.

        Args:
            d (dict): dictionary represenation.

        Returns:
            beep.structure.ProcessedCyclerRun: deserialized ProcessedCyclerRun.
        """
        # no need for original datapaths, as their ref paths should
        # be in the subobjects
        featurizers = [MontyDecoder().process_decoded(f) for f in d["featurizers"]]
        return cls(featurizers)

    @classmethod
    def from_json_file(cls, filename):
        """Load a structured run previously saved to file.

        .json.gz files are supported.

        Loads a BEEPFeatureMatrix from json.

        Can be used in combination with files serialized with BEEPFeatures.to_json_file.

        Args:
            filename (str, Pathlike): a json file from a structured run, serialzed with to_json_file.

        Returns:
            None
        """
        return loadfn(filename)

    def to_json_file(self, filename):
        """Save a BEEPFeatureMatrix to disk as a json.

        .json.gz files are supported.

        Not named from_json to avoid conflict with MSONable.from_json(*)

        Args:
            filename (str, Pathlike): The filename to save the file to.
            omit_raw (bool): If True, saves only structured (NOT RAW) data.
                More efficient for saving/writing to disk.

        Returns:
            None
        """
        d = self.as_dict()
        dumpfn(d, filename)
